local ai = {}
ai.TUNABLE = {}

--- Guess how many castle blocks this player can build.
-- @param r a "resources" struct
-- @return a whole number
function ai.estimate_maximum_contribution(r)

	--FIXME: with 3f,2w,2s,2c then it is possible to make 3 contributions,
	--but this function does not recognize this

	return math.max(
			math.min(r.food, r.wood, r.stone + r.cloth),
			math.min(r.food, r.stone, r.wood + r.cloth),
			math.min(r.food, r.cloth, r.wood + r.stone)
			)
end

--- Choose a move when a builder is being activated.
-- E.g. the carpenter or mason
-- @return a movestr
function ai.select_builder_move()
	local moves_array = all_available_moves()
	if #moves_array >= 2 and moves_array[1] == "noaction" then
		return moves_array[2]
	else
		return moves_array[1]
	end
end

--- Choose a move when the castle is being activated.
-- @return a movestr
function ai.select_castle_move()
	local all_moves = all_available_moves()
	if #all_moves == 1 then
		return all_moves[1]
	end

	local options_by_count = {}
	for i,m in ipairs(all_moves) do
		local c = string.match(m, "^castle (.*)$")
		assert(c)
		if c == "nobuild" then
			--do nothing
		else
			local n = math.floor(string.len(c) / 3)
			assert(n >= 1)
			if not options_by_count[n] then
				options_by_count[n] = {}
			end
			table.insert(options_by_count[n], m)
		end
	end

	local my_best = #options_by_count
	assert(my_best >= 1)

	if state.castle_phase == 3 then
		--contribute as much as possible
		return options_by_count[my_best][1]
	end

	local best_from_followers = 0
	for i = state.turn + 1, #state.buildings[CASTLE_POSITION].occupants do
		local pid = state.buildings[CASTLE_POSITION].occupants[i]
		local m = ai.estimate_maximum_contribution(state.players[pid].resources)
		if m > best_from_followers then
			best_from_followers = m
		end
	end

	if state.turn > 1 and my_best <= state.best_contribution_amount then
		--one of my opponents already gave as many blocks as I can
		--contribute this turn.
		--So... contribute just one block, saving the rest for
		--a future turn.
		return options_by_count[1][1]
	end

	if best_from_followers >= 1 and my_best >= best_from_followers then
		--give just enough so that none of my opponents can beat
		--my contribution
		return options_by_count[best_from_followers][1]
	end

	if state.best_contribution_amount + 1 <= my_best then
		return options_by_count[state.best_contribution_amount + 1][1]
	end

	--give as many as possible
	return options_by_count[my_best][1]
end

--- Choose a valid move when the merchant guild or bridge is being activated.
-- @return a movestr
function ai.select_provost_move()
	local whoami = current_turn()
	local is_free = (state.active_building == MERCHANT_GUILD_POSITION)

	local last_my_worker = 0
	local last_opponent_worker = 0
	local need_to_save_coins = false
	for i = BRIDGE_POSITION + 1, LAST_EMPTY_POSITION do
		local bldg = state.buildings[i]
		if bldg.occupants[1] == whoami then
			last_my_worker = i
			if bldg.ref.id == "lawyer"
				or bldg.ref.id == "bank"
				or bldg.ref.id == "church"
				or bldg.ref.id == "wood-peddler"
				or bldg.ref == "fixed-peddler"
			then
				if i <= state.provost then
					need_to_save_coins = true
				end
			end
		elseif bldg.occupants[1] then
			last_opponent_worker = i
		end
	end
	if is_free then
		if last_my_worker > state.bailiff - 3 then
			return "provost 3"
		else
			return "provost -3"
		end
	else
		local spare_coins = state.players[whoami].resources.money - 2
		if need_to_save_coins then
			spare_coins = spare_coins - 2
		end
		spare_coins = math.min(3, math.max(0, spare_coins))

		if spare_coins == 0 then
			return "noaction"
		end

		local i_am_last_on_bridge = (state.turn == #state.order)
		if last_my_worker > state.provost then
			if i_am_last_on_bridge then
				spare_coins = math.min(spare_coins, last_my_worker - state.provost)
			end
			return string.format("provost %d", spare_coins)
		elseif last_my_worker < state.provost - 3
		then
			local steps = spare_coins
			steps = math.min(steps, state.provost - (BRIDGE_POSITION + 1))
			while steps >= 1 and not state.buildings[state.provost + 1 - steps].occupants[1] do
				steps = steps - 1
			end
			if steps >= 1 then
				return string.format("provost %d", -steps)
			else
				return "noaction"
			end
		else
			return "noaction"
		end
	end
end

function ai.select_move()
	ai.context = {}
	local ph = current_phase()
	if ph == PHASE_WORKER_PLACEMENT then
		return ai.select_worker_placement()
	elseif ph == PHASE_MOVE_PROVOST then
		return ai.select_provost_move()
	elseif ph == PHASE_FAVOR then
		return ai.select_favor_move()
	else
		local whoami = current_turn()
		if state.active_building > 0 and state.active_building ~= BRIDGE_POSITION then
			ai.context.position = state.active_building
			local bldgref = state.buildings[state.active_building].ref
			if bldgref == LAWYER
				or bldgref == ARCHITECT1
				or bldgref == ARCHITECT2
				or bldgref == CARPENTER1
				or bldgref == CARPENTER2
				or bldgref == MASON
			then
				return ai.select_builder_move()
			elseif bldgref == JOUST_FIELD then
				return ai.select_favor_move()
			elseif bldgref == CASTLE then
				return ai.select_castle_move()
			end

			if bldgref == ALCHEMIST and ai.select_alchemist_move then
				return ai.select_alchemist_move()
			elseif bldgref == BANK and ai.select_bank_move then
				return ai.select_bank_move()
			elseif bldgref == CHURCH and ai.select_church_move then
				return ai.select_church_move()
			elseif bldgref == JEWELER and ai.select_jeweler_move then
				return ai.select_jeweler_move()
			elseif bldgref == TAILOR and ai.select_tailor_move then
				return ai.select_tailor_move()
			end

			return ai.select_worker_action()
		else
			error(string.format("phase=%d, active building=%d",
				state.phase, state.active_building))
		end
	end
end

-- Determines the marginal value of some resources (can be negative)
-- in comparison to the player's current resources.
--
-- Usage: score = ai:value_of_resources_delta(bag)
--     @param bag the net gain/loss of resources
--
-- Usage: score = ai:value_of_resources_delta(bag, r)
--     @param bag the net gain/loss of resources
--     @param r the resources on hand
--
function ai.value_of_resources_delta(ai, bag, orig_resources)
	local whoami = current_turn()
	local me = state.players[whoami]
	orig_resources = orig_resources or me.resources
	local adj_r = {
		food = orig_resources.food + bag.food,
		wood = orig_resources.wood + bag.wood,
		stone = orig_resources.stone + bag.stone,
		cloth = orig_resources.cloth + bag.cloth,
		gold = orig_resources.gold + bag.gold,
		money = orig_resources.money + bag.money,
		}
	local my_score = ai.value_of_resources(adj_r)
	local orig_score = ai.value_of_resources(orig_resources)

	assert(my_score >= 0 and my_score < math.huge, "value of '" .. resource_bag.as_string(adj_r) .. "' is invalid: " .. my_score)
	assert(orig_score >= 0 and orig_score < math.huge)

	return my_score - orig_score
end

function ai.pick_building_owner_action(ai, bldgref)
	if bldgref.id == "inn" then
		return "noaction"
	end
	assert(bldgref.owner_actions and bldgref.owner_actions[1])

	local best_move
	local best_score = -math.huge

	for i,movestr in ipairs(bldgref.owner_actions) do
		
		local rsrc_str = string.match(movestr, "^take (.*)$")
		if rsrc_str then
			local bag = resource_bag.parse(rsrc_str)
			local my_score = ai:value_of_resources_delta(bag)
			if my_score > best_score then
				best_score = my_score
				best_move = movestr
			end
		end
	end

	assert(best_move)
	return best_score,best_move
end

function ai.pick_building_pickup_action(ai, bldgref)

	if not (bldgref.actions and bldgref.actions[1]) then
		return 0,"noaction"
	end

	local best_move = "noaction"
	local best_score = 0

	for i,movestr in ipairs(bldgref.actions) do

		local rsrc_str = string.match(movestr, "^take (.*)$")
		if rsrc_str then
			local bag = resource_bag.parse(rsrc_str)
			local my_score = ai:value_of_resources_delta(bag)
			if my_score > best_score then
				best_score = my_score
				best_move = movestr
			end
		end
	end

	return best_score,best_move
end

return ai
