local ai = {}
ai.valuation = load_ai_personality("valuation_nn.inc")
ai.quickmove = load_ai_personality("quickmove1.inc")

local AI_SEARCH_BREADTH = 3
local AI_SEARCH_DEPTH = 1
ai.VERBOSE = false
ai.WEIGHTS = ai.valuation.WEIGHTS
ai.USE_FUZZY_SCORING = false

--this variable gets reset between calls to select_move()
local context = {}

local function debug(message)
	if ai.VERBOSE then
		print(message)
	end
end

ai.estimate_rounds_left = ai.valuation.estimate_num_rounds_left
ai.help_learn           = ai.valuation.help_learn
ai.estimate_player_final_scores = ai.valuation.estimate_player_final_scores

local function dump_choice(choice)
	local moves_string = choice.move
	local ee = choice.parent
	while ee.move do
		moves_string = ee.move .. "|" .. moves_string
		ee = ee.parent
	end
	if choice.principal_variation then
		moves_string = moves_string .. "   ("
		for i,m in ipairs(choice.principal_variation) do
			if i > 1 then moves_string = moves_string .. "|" end
			moves_string = moves_string .. m
		end
		moves_string = moves_string .. ")"
	end
	local MAX_LINE_LENGTH = 800
	if string.len(moves_string) > (MAX_LINE_LENGTH - 6) then
		moves_string = string.sub(moves_string, 1, (MAX_LINE_LENGTH - 8)) .. ".."
	end
	local my_score = ai.valuation.evaluate_scores(choice.estimated_scores, context.whoami)
	print(string.format("%5.1f [%5.1f] %s", my_score, choice.estimated_scores[context.whoami], moves_string))
end

local function new_node(parent)
	context.count_nodes = context.count_nodes + 1
	local choices = caylus.is_end_of_game(state) and {}
			or clone1(all_available_moves())
	return {
		state = state,
		alternate_moves = choices,
		evaluated = {},
		preferred_moves = {},
		estimated_scores = nil,
		parent = parent,
		turn = (#choices > 1 and current_turn() or 1),
		}
end

--check_alternatives(n)
--  n: a node of the game tree
--  returns: true if an alternative was found to be better
--     false otherwise
--
-- Creates a PV for each alternate choice at node n,
-- and updates node's n "preferred_moves" etc to reflect the
-- evaluated paths, and updates ancestors preferred moves
-- if necessary.
--
local function check_alternatives(n)

	if #n.alternate_moves == 0 then
		return false
	end

	local prior_best = n.preferred_moves[1]
	ai.evaluate_node(n, math.huge, 0)
	if prior_best == n.preferred_moves[1] then
		return false
	end

	if ai.VERBOSE then
		print(string.format("pv changed %s to %s",
				prior_best.move, n.preferred_moves[1].move))
	end
	return true
end

local function too_much_thinking()
	return context.count_nodes > 2000
	--return is_time_exceeded()
end

local evaluate_scores = ai.valuation.evaluate_scores
local function get_comparer(node)
	local function comparer(lhs,rhs)
		local lhs_score = evaluate_scores(lhs.estimated_scores, node.turn)
		local rhs_score = evaluate_scores(rhs.estimated_scores, node.turn)
		return lhs_score > rhs_score
	end
	return comparer
end

local function all_child_nodes_changed(node)

	local prior_best = assert(node.preferred_moves[1])
	local comparer = get_comparer(node)

	table.sort(node.preferred_moves, comparer)
	if prior_best ~= node.preferred_moves[1] then
		if ai.VERBOSE then
			print(string.format("pv changed %s to %s",
					prior_best.move, node.preferred_moves[1].move))
		end
	end

	node.estimated_scores = node.preferred_moves[1].estimated_scores
	if node.turn ~= context.whoami and ai.USE_FUZZY_SCORING then
		node.estimated_scores = clone1(node.preferred_moves[1].estimated_scores)
		local basis = evaluate_scores(node.preferred_moves[1].estimated_scores, node.turn)
		local total_weight = 1
		for i=2,#node.preferred_moves do
			local est_score = node.preferred_moves[i].estimated_scores
			local b = evaluate_scores(est_score, node.turn)
			assert(basis >= b)
			local w = 1 / (1 + (basis - b))
			for j,s in pairs(est_score) do
				node.estimated_scores[j] = node.estimated_scores[j] + w * s
			end
			total_weight = total_weight + w
		end
		for j,s in pairs(node.estimated_scores) do
			node.estimated_scores[j] = s / total_weight
		end
	end

	return true
end

local function child_node_changed(node, idx)

	--TODO- since we know only one node's score changed,
	--perhaps a more optimal algorithm can be used.

	return all_child_nodes_changed(node)

end

local expand_alternative
local create_principal_variation

function expand_alternative(node,next_move)
	context.count_clones = context.count_clones + 1
	state = clone_game_state(assert(node.state))
	do_move(next_move)
	local child_node = new_node(node)
	child_node.move = next_move
	create_principal_variation(child_node)
	node.evaluated[next_move] = child_node

	--FIXME- better to just insert at the correct spot
	table.insert(node.preferred_moves, child_node)
	return child_node_changed(node, #node.preferred_moves)
end

local function refine_building_activation_plan(node)

	if not node.preferred_moves[1] then
		return false
	end

	if node.state.phase == PHASE_WORKER_PLACEMENT then
		return refine_building_activation_plan(node.preferred_moves[1])
			and child_node_changed(node, 1)
	end

	local n = node
	if node.state.phase == PHASE_ACTIVATE_BUILDINGS
		and not too_much_thinking()
	then

		if #node.alternate_moves > 0 then
			if ai.VERBOSE then
				print(string.format("examining what to do when %s (%d) is activated",
					n.state.buildings[n.state.active_building].ref.name,
					n.state.active_building))
			end
			if check_alternatives(node) then
				return true
			end
		end

		return refine_building_activation_plan(node.preferred_moves[1])
			and child_node_changed(node, 1)
	end
	return false
end

local function search_tree_for_preferred_move(node)
	if node.state.phase ~= PHASE_ACTIVATE_BUILDINGS then
		return nil
	end

	local n = context.root_node
	while n.preferred_moves[1]
			and n.state.phase == PHASE_WORKER_PLACEMENT do
		n = n.preferred_moves[1]
	end
	while n.preferred_moves[1]
			and n.state.phase == PHASE_ACTIVATE_BUILDINGS
			and n.state.active_building < node.state.active_building do
		n = n.preferred_moves[1]
	end

	while n.preferred_moves[1]
			and n.state.phase == PHASE_ACTIVATE_BUILDINGS
			and n.state.active_building == node.state.active_building
			and n.state.turn < node.state.turn do
		n = n.preferred_moves[1]
	end

	if n.preferred_moves[1]
			and n.state.phase == PHASE_ACTIVATE_BUILDINGS
			and n.state.active_building == node.state.active_building
			and n.state.turn == node.state.turn
	then
		return n.preferred_moves[1].move
	else
		return nil
	end
end

local function use_heuristics_to_order_alternates(node)

	local function analyze_this_move_next(movestr)
		for i,m in ipairs(node.alternate_moves) do
			if m == movestr then
				local tmp = node.alternate_moves[#node.alternate_moves]
				node.alternate_moves[#node.alternate_moves] = node.alternate_moves[i]
				node.alternate_moves[i] = tmp
				return true
			end
		end
		return false
	end

	local movestr
	--first, we use the search tree to find what move has been
	--favored at this level from another branch of the game tree

	movestr = search_tree_for_preferred_move(node)
	if movestr then
		if analyze_this_move_next(movestr) then
			return
		end
	end

	--that didn't work, so now we use our "quick_ai" to come up
	--with a preferred move
	movestr = ai.quickmove.select_move()
	analyze_this_move_next(movestr)
	return
end

local function node_has_psuedo_pass_ancestor(node)
	if node.move == "~pass" then
		return true
	elseif node.parent then
		return node_has_psuedo_pass_ancestor(node.parent)
	else
		return false
	end
end

function create_principal_variation(node)

	if node.preferred_moves[1] then
		return false
	end

	state = node.state
	if caylus.is_end_of_game(state) or context.is_stable_state() then
		--new round; we can evaluate here
		if not node.estimated_scores then
			node.estimated_scores = ai.valuation.evaluate_position_at_round_end(context.whoami,
				{},
				{ in_truncated_round = node_has_psuedo_pass_ancestor(node), }
				)
			if ai.VERBOSE then
				dump_choice(node)
				--dump_evaluation_details()
			end
		end
		return false
	end

	assert(#node.preferred_moves == 0)

	--no moves have been evaluated yet
	--apply some heuristics to determine which move
	--we will look at first

	if #node.alternate_moves > 1 then
		use_heuristics_to_order_alternates(node)
	else
		assert(#node.alternate_moves >= 1)
	end

	local next_move = table.remove(node.alternate_moves)
	return expand_alternative(node, next_move)
end

local function skip_to_my_turn_again()
	assert(state.phase == PHASE_WORKER_PLACEMENT)
	assert(not state.players[context.whoami].passed)

	while state.order[state.turn] ~= context.whoami do
		next_turn()
	end
end

local function skip_to_building_activation_phase()

	local old_rule = NO_PASS_REWARD
	NO_PASS_REWARD = true
	while state.phase == PHASE_WORKER_PLACEMENT do
		do_move("pass")
	end
	NO_PASS_REWARD = old_rule
end

local function check_one_alternative_maybe(node,next_move)
	for i=1,#node.alternate_moves do
		if node.alternate_moves[i] == next_move then
			table.remove(node.alternate_moves,i)
			return expand_alternative(node, next_move)
		end
	end
	return false
end

local function report_choices(node)
	assert(ai.VERBOSE)
	
	state = node.state
	local _is_my_turn = is_my_turn(context.whoami)
	print(_is_my_turn and "my choices:" or "opponent choices:")
	for i,n in ipairs(node.preferred_moves) do
		dump_choice(n)
	end
end

local function generate_followup_moves(node,breadth,maxdepth)
	if maxdepth < 0 then
		debug("reached maxdepth")
		return false
	end
	if node.state.phase ~= PHASE_WORKER_PLACEMENT then
		return false
	end

	local child2 = node.preferred_moves[1]
	if not child2 then return false end

	debug("checking for alternatives to "..child2.move)

	if check_alternatives(child2) then
		local comparer = get_comparer(node)
		local favored_moves = {}
		favored_moves[child2.move] = 1
		favored_moves[child2.preferred_moves[1].move] = 1
		for i = 2, #node.preferred_moves do
			local child3 = node.preferred_moves[i]
			if comparer(child2, child3) then break end
			local child4 = child3.preferred_moves[1]
			for m in pairs(favored_moves) do
				check_one_alternative_maybe(child3, m)
			end
			if child3.preferred_moves[1] ~= child4 then
				--TODO
			end
			if i <= breadth then
				favored_moves[child3.move] = 1
			end
		end
		if ai.VERBOSE then
			report_choices(node)
		end
		return all_child_nodes_changed(node)
	end

	return false
end

function ai.evaluate_node(node, breadth, depth)

	assert(context.whoami)
	assert(context.is_stable_state)

	state = node.state

	if caylus.is_end_of_game(state) or context.is_stable_state() then
		--new round; we can evaluate here
		if not node.estimated_scores then
			node.estimated_scores = ai.valuation.evaluate_position_at_round_end(context.whoami,
					{},
					{ in_truncated_round = node_has_psuedo_pass_ancestor(node), }
					)

			if ai.VERBOSE then
				dump_choice(node)
				--dump_evaluation_details()
			end
		end
		return node.estimated_scores, nil
	end

	--we maximize for the player whose turn it is

	if #node.preferred_moves == 0 then

		--no moves have been evaluated yet
		--apply some heuristics to determine which move
		--we will look at first

		if #node.alternate_moves > 1 then
			use_heuristics_to_order_alternates(node)
		else
			assert(#node.alternate_moves >= 1)
		end
	end

	if #node.alternate_moves == 0 then

		--all moves have been evaluated
		return node.estimated_score, node.preferred_moves[1].move
	end

	while #node.preferred_moves < breadth
		and #node.alternate_moves > 0
		and not (#node.preferred_moves >= 1 and too_much_thinking())
	do

		local next_move

		next_move = table.remove(node.alternate_moves)
		expand_alternative(node, next_move)

	end --loop through available choices at this node

	if ai.VERBOSE and #node.preferred_moves > 1 then
		report_choices(node)
	end

	state = node.state
	if depth >= 1 then

		--check that the "principal variation" we're using to
		--determine the valuation of the move is really the
		--best sequence of moves

		--if we find the principal variation changes, we
		--repeat the process with the adjusted variation until
		--we settle on an accurate variation

		local function refine_first_alternates_pv()
			if #node.preferred_moves >= 2
				and refine_building_activation_plan(node.preferred_moves[2])
			then
				return child_node_changed(node, 2)
			end
			return false
		end

		while not too_much_thinking()
		do

			if not (
				refine_building_activation_plan(node) or
				--refine_first_alternates_pv() or
				generate_followup_moves(node,3,1)
				) then

				break
			end

		end --loop
	end

	if ai.VERBOSE and #node.preferred_moves > 1 then
		report_choices(node)
	end

	local best = node.preferred_moves[1]
	node.estimated_scores = best.estimated_scores
	return node.estimated_scores, best.move

end

function ai.select_move()

	start_thinking()
	context = {
		whoami = current_turn(),
		count_nodes = 0,
		count_clones = 0,
		}
	local saved_state = state

	if state.phase == PHASE_SCORING then
		context.is_stable_state = function()
				return state.round > saved_state.round
			end
	else
		context.is_stable_state = function()
				return state.phase == PHASE_SCORING
					or state.round > saved_state.round
			end
	end

	context.root_node = new_node(nil)

	local next_move

	if #context.root_node.alternate_moves == 1 then
		next_move = context.root_node.alternate_moves[1]
	else
		ai.evaluate_node(context.root_node, math.huge, 1);
		next_move = context.root_node.preferred_moves[1].move
	end

	--finish_thinking()
	--print(string.format("%.2fs / %d nodes / %s", get_elapsed(), context.count_nodes, next_move))
	if ai.VERBOSE then
		print(string.format("%d nodes", context.count_nodes))
		print(string.format("%d clones", context.count_clones))
	end

	state = saved_state
	if not ai.SAVE_ANALYSIS then
		context = nil
		collectgarbage("collect")
	end

	return next_move

end

function ai.show_move_evaluation(movestr)
	local saved_state = state
	local node = context.root_node
	local moves = {}
	for next_move in movestr:split(";") do
		if not node.evaluated[next_move] then
			expand_alternative(node, next_move)
		end
		node = assert(node.evaluated[next_move])
		table.insert(moves, next_move)
	end

	--follow principal variation to end
	while node.preferred_moves[1] do
		node = node.preferred_moves[1]
		table.insert(moves, node.move)
	end

	dump_choice(node)
	state = node.state
	ai.valuation.evaluate_position_detailed()

	state = saved_state
	return node.state, moves
end

return ai
